/*
 * RFCNB (RFC1001/RFC1002) NetBIOS implementation
 * 
 * Version 1.0
 * SMB/RFCNB Utility Routines ...
 * 
 * Copyright (C) Richard Sharpe 1996
 * 
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"


static const char *SMB_Protocols[] = {
                  "PC NETWORK PROGRAM 1.0",
                  "MICROSOFT NETWORKS 1.03",
                  "MICROSOFT NETWORKS 3.0",
                  "DOS LANMAN1.0",
                  "LANMAN1.0",
                  "DOS LM1.2X002",
                  "LM1.2X002",
                  "DOS LANMAN2.1",
                  "LANMAN2.1",
                  "Samba",
                  "NT LM 0.12",
                  "NT LANMAN 1.0",
                  NULL
                };

static const int SMB_Types[] = {
                 SMB_P_Core,
                 SMB_P_CorePlus,
                 SMB_P_DOSLanMan1,
                 SMB_P_DOSLanMan1,
                 SMB_P_LanMan1,
                 SMB_P_DOSLanMan2,
                 SMB_P_LanMan2,
                 SMB_P_LanMan2_1,
                 SMB_P_LanMan2_1,
                 SMB_P_NT1,
                 SMB_P_NT1,
                 SMB_P_NT1,
                 SMB_P_Unknown
               };

/*
 * Convert a name and pad to 16 chars ("CA") as needed.
 * name1 is a C string with null termination,
 * name2 holds result of conversion.
 */
void RFCNB_CvtPad_Name (const char *name1, char *name2)
{
  char c1, c2;
  int  i, len = strlen (name1);

  for (i = 0; i < 16; i++)
  {
    if (i >= len)
    {
      c1 = 'C';
      c2 = 'A';       /* CA is a space */
    }
    else
    {
      char c = name1[i];
      c1 = (int)(c / 16) + 'A';
      c2 = (int)(c % 16) + 'A';
    }
    name2[i*2]   = c1;
    name2[i*2+1] = c2;
  }
  name2[32] = '\0';
}

/*
 * Converts an ASCII NetBios Name (16 chars) to an RFCNB Name (32 chars)
 * Uses the encoding in RFC1001. Each nibble of byte is added to 'A'
 * to produce the next byte in the name.
 * 
 * This routine assumes that aName is 16 bytes long and that nbName has
 * space for 32 chars, so be careful.
 */
void RFCNB_AName_To_NBName (const char *aName, char *nbName)
{
  char c, c1, c2;
  int  i;

  for (i = 0; i < 16; i++)
  {
    c  = aName[i];
    c1 = (c >> 4) + 'A';
    c2 = (c & 0xF) + 'A';

    nbName[i*2]   = c1;
    nbName[i*2+1] = c2;
  }
  nbName[32] = '\0';
}

/*
 * Do the reverse of the above
 */
void RFCNB_NBName_To_AName (const char *nbName, char *aName)
{
  int i;

  for (i = 0; i < 16; i++)
  {
    char c1  = nbName[i*2];
    char c2  = nbName[i*2+1];
    aName[i] = 16 * ((int)c1 - 'A') + ((int)c2 - 'A');
  }
  aName[i] = '\0';
}

/*
 * Print a string of bytes in HEX etc
 */
void RFCNB_Print_Hex (FILE *fd, const struct RFCNB_Pkt *pkt,
                      int ofs, int len)
{
  static char hexChar[17] = "0123456789ABCDEF";
  char   c1, c2, outbuf1[33];
  BYTE   c;
  int    i, j = 0;

  /* We only want to print as much as specified in len
   */
  while (pkt)
  {
    for (i = 0; i < (len > pkt->len ? pkt->len : len) - ofs; i++)
    {
      c  = pkt->data [i + ofs];
      c1 = hexChar [c >> 4];
      c2 = hexChar [c & 0xF];

      outbuf1 [j++] = c1;
      outbuf1 [j++] = c2;

      if (j == 32)    /* reached right column */
      {
        outbuf1 [j] = '\0';
        fprintf (fd, "    %s\n", outbuf1);
        j = 0;
      }
    }
    ofs  = 0;
    len -= pkt->len;
    pkt  = pkt->next;
  }

  /* Print last lot in the buffer
   */
  if (j > 0)
  {
    outbuf1[j] = '\0';
    fprintf (fd, "    %s\n", outbuf1);
  }
  fputc ('\n', fd);
}

/*
 * Allocate a packet with x bytes of data.
 */
struct RFCNB_Pkt *RFCNB_Alloc_Pkt (int data_size)
{
  struct RFCNB_Pkt *pkt = calloc (sizeof(*pkt), 1);

  if (!pkt)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    return (NULL);
  }

  pkt->next = NULL;
  pkt->len  = data_size;

  if (data_size == 0)
     return (pkt);

  pkt->data = calloc (data_size, 1);
  if (!pkt->data)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    free (pkt);
    return (NULL);
  }
  return (pkt);
}

/*
 * Free up a packet
 */
void RFCNB_Free_Pkt (struct RFCNB_Pkt *pkt)
{
  struct RFCNB_Pkt *pkt_next;
  char  *data_ptr;

  while (pkt)
  {
    pkt_next = pkt->next;
    data_ptr = pkt->data;
    if (data_ptr)
       free (data_ptr);
    free (pkt);
    pkt = pkt_next;
  }
}

/*
 * Print an RFCNB packet
 */
void RFCNB_Print_Pkt (FILE *fd, const char *in_out,
                      const struct RFCNB_Pkt *pkt, int len)
{
  char lname[17];
  int  data_len = 40;
  BOOL sending = (strcmp(in_out,"sent") == 0);

  /* We assume that the first fragment is the RFCNB Header
   * We should loop through the fragments printing them out
   */
  fprintf (fd, "RFCNB Pkt %s: ", in_out);

  switch (RFCNB_Pkt_Type (pkt->data))
  {
    case RFCNB_SESSION_MESSAGE:
         fprintf (fd, "SESSION MESSAGE: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));

#ifdef RFCNB_PRINT_DATA
         data_len = RFCNB_Pkt_Len (pkt->data) - RFCNB_Pkt_Hdr_Len;
#endif
         RFCNB_Print_Hex (fd, pkt, RFCNB_Pkt_Hdr_Len, data_len);

         if (Prot_Print_Routine)  /* Print the rest of the packet */
           (*Prot_Print_Routine) (fd, !sending, pkt, RFCNB_Pkt_Hdr_Len,
                                  RFCNB_Pkt_Len(pkt->data) - RFCNB_Pkt_Hdr_Len);
         break;

    case RFCNB_SESSION_REQUEST:
         fprintf (fd, "SESSION REQUEST: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));
         RFCNB_NBName_To_AName ((char*)(pkt->data + RFCNB_Pkt_Called_Offset),
                                 lname);
         fprintf (fd, "  Called Name: %s\n", lname);
         RFCNB_NBName_To_AName ((char*)(pkt->data + RFCNB_Pkt_Calling_Offset),
                                lname);
         fprintf (fd, "  Calling Name: %s\n", lname);
         break;

    case RFCNB_SESSION_ACK:
         fprintf (fd, "RFCNB SESSION ACK: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));
         break;

    case RFCNB_SESSION_REJ:
         fprintf (fd, "RFCNB SESSION REJECT: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));

         if (RFCNB_Pkt_Len(pkt->data) < 1)
              fprintf (fd, "   Protocol Error, short Reject packet!\n");
         else fprintf (fd, "   Error = %x\n", CVAL (pkt->data, RFCNB_Pkt_Error_Offset));
         break;

    case RFCNB_SESSION_RETARGET:
         fprintf (fd, "RFCNB SESSION RETARGET: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));

         /* !!to-do: print out the IP address etc and the port? */
         break;

    case RFCNB_SESSION_KEEP_ALIVE:
         fprintf (fd, "RFCNB SESSION KEEP ALIVE: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));
         break;

    default:
         break;
  }
}

/*
 * Resolve a name into an address
 */
int RFCNB_Name_To_IP (const char *host, struct in_addr *dest_IP)
{
  DWORD addr = inet_addr (host);

  if (addr == INADDR_NONE)  /* not a dotted IP-address */
  {
    /* !!to-do: Try NetBIOS name lookup first.
     *          Use LMHOSTS-file, HOSTS-file, Wins and DNS last.
     */
    struct hostent *hp = gethostbyname (host);

    if (!hp)
    {
      RFCNB_errno = RFCNBE_BadName;
      RFCNB_saved_errno = errno;
      return (-1);
    }
    dest_IP->s_addr = *(u_long*)&hp->h_addr;
  }
  else
    dest_IP->s_addr = addr;
  return (0);
}

/*
 * Disconnect the TCP connection to the server
 */
int RFCNB_Close (int socket)
{
  SOCKET_CLOSE (socket);

  /* If we want to do error recovery, here is where we put it
   */
  return (0);
}

/*
 * Connect to the server specified in the IP address.
 * Not sure how to handle socket options etc.
 */
int RFCNB_IP_Connect (struct in_addr dest_IP, int port)
{
  struct sockaddr_in addr;
  int    fd = socket (AF_INET, SOCK_STREAM, 0);

  if (fd < 0)
  {
    RFCNB_errno = RFCNBE_BadSocket;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  bzero (&addr, sizeof(addr));
  memcpy (&addr.sin_addr, &dest_IP, sizeof(dest_IP));

  addr.sin_port   = htons (port);
  addr.sin_family = AF_INET;

  /* Now connect to the destination. bind() is implicit which selects
   * a free local port for us.
   */
  if (connect (fd, (struct sockaddr*)&addr, sizeof(addr)) < 0)
  {
    SOCKET_CLOSE (fd);
    RFCNB_errno = RFCNBE_ConnectFailed;
    RFCNB_saved_errno = errno;
    return (-1);
  }
  return (fd);
}

/*
 * Handle the details of establishing the RFCNB session with
 * remote end
 */
int RFCNB_Session_Req (struct RFCNB_Con *con,
                       const  char      *called_name,
                       const  char      *calling_name,
                       BOOL             *redirect,
                       struct in_addr   *dest_IP,
                       unsigned short   *port)
{
  struct RFCNB_Pkt *pkt, res_pkt;
  char   *sess_pkt;
  char    resp[16];  /* should be >=9 bytes */
  int     len;

  pkt = RFCNB_Alloc_Pkt (RFCNB_Pkt_Sess_Len);
  if (!pkt)
     return (-1);

  sess_pkt = pkt->data;         /* Get pointer to packet proper */

  sess_pkt [RFCNB_Pkt_Type_Offset] = RFCNB_SESSION_REQUEST;
  RFCNB_Put_Pkt_Len (sess_pkt, RFCNB_Pkt_Sess_Len - RFCNB_Pkt_Hdr_Len);
  sess_pkt [RFCNB_Pkt_N1Len_Offset] = 32;
  sess_pkt [RFCNB_Pkt_N2Len_Offset] = 32;

  RFCNB_CvtPad_Name (called_name,  sess_pkt + RFCNB_Pkt_Called_Offset);
  RFCNB_CvtPad_Name (calling_name, sess_pkt + RFCNB_Pkt_Calling_Offset);

  /* Now send the packet
   */
  SMB_debug (2, "Sending packet: ");

  len = RFCNB_Put_Pkt (con, pkt, RFCNB_Pkt_Sess_Len);
  if (len < 0)
     return (-1);

  SMB_debug (2, "Getting packet.\n");

  res_pkt.data = resp;
  res_pkt.len  = sizeof (resp);
  res_pkt.next = NULL;

  len = RFCNB_Get_Pkt (con, &res_pkt, sizeof(resp));
  if (len < 0)
     return (-1);

  /* Now analyze the packet
   */
  switch (RFCNB_Pkt_Type (resp))
  {
    case RFCNB_SESSION_REJ: /* Didn't like us, too bad */

         /* Why did we get rejected ?
          */
         switch (CVAL (resp, RFCNB_Pkt_Error_Offset))
         {
           case 0x80:
                RFCNB_errno = RFCNBE_CallRejNLOCN;
                break;
           case 0x81:
                RFCNB_errno = RFCNBE_CallRejNLFCN;
                break;
           case 0x82:
                RFCNB_errno = RFCNBE_CallRejCNNP;
                break;
           case 0x83:
                RFCNB_errno = RFCNBE_CallRejInfRes;
                break;
           case 0x8F:
                RFCNB_errno = RFCNBE_CallRejUnSpec;
                break;
           default:
                RFCNB_errno = RFCNBE_ProtErr;
                break;
         }
         return (-1);

    case RFCNB_SESSION_ACK: /* Got what we wanted */
         return (0);

    case RFCNB_SESSION_RETARGET: /* Go elsewhere */
         *redirect       = TRUE;
         dest_IP->s_addr = *(DWORD*) &resp[RFCNB_Pkt_IP_Offset];
         *port           = SVAL (resp, RFCNB_Pkt_Port_Offset);
         return (0);

    default:                     /* A protocol error */
         RFCNB_errno = RFCNBE_ProtErr;
         return (-1);
  }
  /* NOT REACHED */
}

/*
 * Print out an SMB pkt in all its gory detail
 */
void SMB_Print_Pkt (FILE *fd, const char *in_out,
                    const struct RFCNB_Pkt *pkt, int len)
{
  /* Well, just how do we do this... print it I suppose
   * 1. Print out the SMB header
   * 2. Print the command
   * 3. Print the other bits in the header
   *
   * etc.
   * steal parser from tcpdump (parse_nbt*.c)
   */
  ARGSUSED (fd);
  ARGSUSED (in_out);
  ARGSUSED (pkt);
  ARGSUSED (len);
}

/*
 * Convert a DOS date/time to a local host type date time for printing
 */
const char *SMB_DOSTimToStr (ULONG dos_time)
{
  static char buf[48];
  static char months[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

  int sec   =  (dos_time & 0x0000001FUL) * 2;
  int min   =  (dos_time & 0x000007E0UL) >> 5;
  int hour  = ((dos_time & 0x0000F800UL) >> 11);

  int day   =  (dos_time & 0x001F0000UL) >> 16;
  int month =  (dos_time & 0x01E00000UL) >> 21;
  int year  = ((dos_time & 0xFE000000UL) >> 25);

  year += (year < 50 ? 2000 : 1980);            /* Y2K hack! */

  sprintf (buf, "%.3s %2d %02d:%02d:%02d %4d",  /* ctime() format */
           months + 3*(month-1), day, hour, min, sec, year);
  return (buf);
}

/*
 * Convert an attribute byte/word etc to a string. We return a pointer
 * to a static string which we guarantee is long enough. If verbose is 
 * true, we print out long form of strings.
 */
const char *SMB_AttrToStr (int attribs, BOOL verbose)
{
  static char buf[128];

  buf[0] = '\0';

  if (attribs & SMB_FA_ROF)
     strcat (buf, (verbose ? "Read Only " : "R"));

  if (attribs & SMB_FA_HID)
     strcat (buf, (verbose ? "Hidden " : "H"));

  if (attribs & SMB_FA_SYS)
     strcat (buf, (verbose ? "System " : "S"));

  if (attribs & SMB_FA_VOL)
     strcat (buf, (verbose ? "Volume " : "V"));

  if (attribs & SMB_FA_DIR)
     strcat (buf, (verbose ? "Directory " : "D"));

  if (attribs & SMB_FA_ARC)
     strcat (buf, (verbose ? "Archive " : "A"));

  return (buf);
}

/*
 * Pick up the Max Buffer Size from the Tree Structure
 */
int SMB_Get_Tree_MBS (struct SMB_Tree *tree)
{
  if (tree)
     return (tree->mbs);
  return (-1);
}

/*
 * Pick up the Max buffer size
 */
int SMB_Get_Max_Buf_Siz (struct SMB_Connect *handle)
{
  if (handle)
     return (handle->max_xmit);
  return (-1);
}

/*
 * Pickup the protocol index from the connection structure
 */
int SMB_Get_Protocol_IDX (struct SMB_Connect *handle)
{
  if (handle)
     return (handle->prot_IDX);
  return (-1);
}

/*
 * Pick up the protocol from the connection structure
 */
int SMB_Get_Protocol (struct SMB_Connect *handle)
{
  if (handle)
     return (handle->protocol);
  return (SMB_P_Unknown);
}

const char * SMB_Get_ProtocolStr (int proto)
{
  static char buf[20];

  if (proto >= 0 && proto < sizeof(SMB_Protocols)/sizeof(char*))
     return SMB_Protocols[proto];

  sprintf (buf, "proto %d??", proto);
  return (buf);
}


/*
 * Figure out what protocol was accepted, given the list of dialect
 * strings we offered, and the index back from the server. We allow
 * for a user supplied list, and assume that it is a subset of our list
 */
static int FigureProtocol (char **table, int prot_index)
{
  int i;

  /* The jobs is easy, just index into table
   */
  if (table == (char**)SMB_Protocols)
     return (SMB_Types[prot_index]);

  /* Search through table looking for a match
   */
  for (i = 0; SMB_Protocols[i]; i++)
      if (!strcmp (table[prot_index], SMB_Protocols[i]))
         return (SMB_Types[i]);
  return (-1);
}
     
/*
 * Negotiate the protocol we will use from the list passed in prots.
 * We return the index of the accepted protocol in NegProt, -1 indicates
 * none acceptible, and our return value is 0 if ok, -1 if problems.
 */
int SMB_Negotiate (struct SMB_Connect *handle, char **prots)
{
  struct RFCNB_Pkt *pkt;
  int    prots_len, i, pkt_len, prot, alloc_len;
  char  *p;

  /* Figure out how long the prot list will be and allocate space for it
   */
  for (i = prots_len = 0; prots[i]; i++)
     prots_len += strlen (prots[i]) + 2;

  /* The -1 accounts for the one byte smb_buf we have because some systems
   * don't like char msg_buf[]
   */
  pkt_len = SMB_negp_len + prots_len;

  /* Make sure that the pkt len is long enough for the max response.
   * Which is a problem, because the encryption key len eec may be long
   */
  if (pkt_len < SMB_hdr_wct_offset + (19 * 2) + 40)
       alloc_len = SMB_hdr_wct_offset + (19 * 2) + 40;
  else alloc_len = pkt_len;

  pkt = RFCNB_Alloc_Pkt (alloc_len);
  if (!pkt)
  {
    SMBlib_errno = SMBlibE_NoSpace;
    return (-1);
  }

  /* Now plug in the bits we need
   */
  bzero (SMB_Hdr (pkt), SMB_negp_len);
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBnegprot;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 0;

  SSVAL (SMB_Hdr (pkt), SMB_negp_bcc_offset, prots_len);

  /* Now copy the prot strings in with the right stuff
   */
  p = (char*) (SMB_Hdr (pkt) + SMB_negp_buf_offset);

  for (i = 0; prots[i]; i++)
  {
    *p++ = SMBdialectID;
    strcpy (p, prots[i]);
    p += strlen (prots[i]) + 1;
  }

  /* Now send the packet and sit back
   */
  if (RFCNB_Send (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending negotiate protocol\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_SendFailed);  /* Failed, check errno */
    return (-1);
  }

  /* Now get the response
   */
  if (RFCNB_Recv (handle->transConnect, pkt, alloc_len) < 0)
  {
    SMB_debug (1, "Error receiving response to negotiate\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_RecvFailed);  /* Failed, check errno */
    return (-1);
  }

  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_debug (1, "SMB_Negotiate fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (-1);
  }

  if (SVAL (SMB_Hdr (pkt), SMB_negrCP_idx_offset) == 0xFFFF)
  {
    SMB_debug (1, "None of our protocols was accepted..");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_NegNoProt;
    return (-1);
  }

  /* Now, unpack the info from the response, if any and evaluate the proto
   * selected. We must make sure it is one we like
   */

  handle->prot_IDX = prot = SVAL (SMB_Hdr(pkt), SMB_negrCP_idx_offset);
  handle->protocol = FigureProtocol (prots, prot);

  if (handle->protocol == SMB_P_Unknown)
  {
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_ProtUnknown;
    return (-1);
  }

  switch (CVAL (SMB_Hdr (pkt), SMB_hdr_wct_offset))
  {
    case 0x01:              /* No more info */
         break;

    case 13:                /* Up to and including LanMan 2.1 */
         handle->security = SVAL (SMB_Hdr (pkt), SMB_negrLM_sec_offset);
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);

         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = SVAL (SMB_Hdr (pkt), SMB_negrLM_mbs_offset);
         handle->maxMPX     = SVAL (SMB_Hdr (pkt), SMB_negrLM_mmc_offset);
         handle->maxVC      = SVAL (SMB_Hdr (pkt), SMB_negrLM_mnv_offset);
         handle->rawSupport = SVAL (SMB_Hdr (pkt), SMB_negrLM_rm_offset);
         handle->sessionKey = IVAL (SMB_Hdr (pkt), SMB_negrLM_sk_offset);
         handle->serverTZ   = SVAL (SMB_Hdr (pkt), SMB_negrLM_stz_offset);
         handle->encrKeyLen = SVAL (SMB_Hdr (pkt), SMB_negrLM_ekl_offset);

         p = (SMB_Hdr(pkt) + SMB_negrLM_buf_offset + handle->encrKeyLen);

         strncpy (p, handle->serverPDom, sizeof(handle->serverPDom)-1);
         break;

    case 17:                /* NT LM 0.12 and LN LM 1.0 */
         handle->security = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_sec_offset);
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);

         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_mbs_offset);
         handle->maxMPX     = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_mmc_offset);
         handle->maxVC      = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_mnv_offset);
         handle->maxRaw     = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_mrs_offset);
         handle->sessionKey = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_sk_offset);
         handle->serverTZ   = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_stz_offset);
         handle->encrKeyLen = CVAL (SMB_Hdr (pkt), SMB_negrNTLM_ekl_offset);

         p = (SMB_Hdr (pkt) + SMB_negrLM_buf_offset + handle->encrKeyLen);

         strncpy (p, handle->serverPDom, sizeof(handle->serverPDom)-1);
         break;

    default:
         SMB_debug (1, "Unknown NegProt response format, ignored\n"
                       "  wct = %i\n", CVAL (SMB_Hdr(pkt), SMB_hdr_wct_offset));
         break;
  }

  SMB_debug (2, "Protocol selected is: %d:%.30s\n", prot, prots[prot]);

  RFCNB_Free_Pkt (pkt);
  return (0);
}

/*
 * Send a TCON to the remote server
 */
struct SMB_Tree *SMB_TreeConnect (struct SMB_Connect *handle,
                                  struct SMB_Tree    *tree,
                                  const  char        *path,
                                  const  char        *password,
                                  const  char        *device)
{
  struct RFCNB_Pkt *pkt;
  struct SMB_Tree  *tree0;
  int    param_len, pkt_len;
  char  *p;

  /* Figure out how much space is needed for path, password, dev
   */
  if (!path || !password || !device)
  {
    SMB_debug (1, "Bad parameter passed to SMB_TreeConnect\n");
    SMBlib_errno = SMBlibE_BadParam;
    return (NULL);
  }

  /* The + 2 is because of the \0 and the marker
   */
  param_len = strlen (path) + 2 +
              strlen (password) + 2 +
              strlen (device) + 2;

  /* The -1 accounts for the one byte smb_buf we have because some systems
   * don't like char msg_buf[]
   */
  pkt_len = SMB_tcon_len + param_len;
  pkt     = RFCNB_Alloc_Pkt (pkt_len);
  if (!pkt)
  {
    SMBlib_errno = SMBlibE_NoSpace;
    return (NULL);
  }

  /* Now allocate a tree for this to go into
   */
  if (!tree)
  {
    tree0 = calloc (sizeof(*tree), 1);
    if (!tree0)
    {
      RFCNB_Free_Pkt (pkt);
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }
  else
  {
    tree0 = tree;
    tree->next = tree->prev = NULL;
  }

  tree0->con = handle;
  strncpy (tree0->path,        path,   sizeof(tree0->path));
  strncpy (tree0->device_type, device, sizeof(tree0->device_type));

  /* Now plug in the values
   */
  bzero (SMB_Hdr (pkt), SMB_tcon_len);
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtcon;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 0;

  SSVAL (SMB_Hdr (pkt), SMB_tcon_bcc_offset, param_len);

  /* Now copy the param strings in with the right stuff
   */
  p = (char *) (SMB_Hdr (pkt) + SMB_tcon_buf_offset);

  *p++ = SMBasciiID;
  strcpy (p, path);
  p += strlen (path) + 1;

  *p++ = SMBasciiID;
  strcpy (p, password);
  p += strlen (password) + 1;

  *p++ = SMBasciiID;
  strcpy (p, device);

  /* Now send the packet and sit back
   */
  if (RFCNB_Send (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending TCon request\n");

    if (!tree)
       free (tree0);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_SendFailed);
    return (NULL);
  }

  /* Now get the response
   */
  if (RFCNB_Recv (handle->transConnect, pkt, pkt_len) < 0)
  {  
    SMB_debug (1, "Error receiving response to TCon\n");

    if (!tree)
      free (tree0);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_RecvFailed);
    return (NULL);
  }

  /* Check out the response type
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_debug (1, "SMB_TCon fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    if (!tree)
       free (tree0);
    SMBlib_SMB_Error = IVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (NULL);
  }

  tree0->tid = SVAL (SMB_Hdr(pkt), SMB_tconr_tid_offset);
  tree0->mbs = SVAL (SMB_Hdr(pkt), SMB_tconr_mbs_offset);

  SMB_debug (2, "TConn succeeded, with TID=%i, Max Xmit=%ld\n",
             tree0->tid, tree0->mbs);

  /* Now link the Tree to the Server Structure
   */
  if (!handle->first_tree)
  {
    handle->first_tree = tree0;
    handle->last_tree  = tree0;
  }
  else
  {
    handle->last_tree->next = tree0;
    tree0->prev = handle->last_tree;
    handle->last_tree = tree0;
  }

  RFCNB_Free_Pkt (pkt);
  return (tree0);
}

int SMB_TreeDisconnect (struct SMB_Tree *tree, BOOL discard)
{
  int    pkt_len        = SMB_tdis_len;
  struct RFCNB_Pkt *pkt = RFCNB_Alloc_Pkt (pkt_len);

  if (!pkt)
  {
    SMBlib_errno = SMBlibE_NoSpace;
    return (-1);
  }

  /* Now plug in the values
   */
  bzero (SMB_Hdr (pkt), SMB_tdis_len);
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtdis;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, tree->con->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, tree->con->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, tree->con->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 0;

  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, tree->tid);
  SSVAL (SMB_Hdr (pkt), SMB_tcon_bcc_offset, 0);

  /* Now send the packet and sit back
   */
  if (RFCNB_Send (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending TDis request\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_SendFailed);
    return (-1);
  }

  /* Now get the response
   */
  if (RFCNB_Recv (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error receiving response to TCon\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = -(SMBlibE_RecvFailed);
    return (-1);
  }

  /* Check out the response type
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_debug (1, "SMB_TDis fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (-1);
  }

  tree->tid = 0xFFFF;    /* Invalid TID */
  tree->mbs = 0;         /* Invalid     */

  SMB_debug (2, "Tree disconnect successful...\n");

  if (discard)
  {
    /* Unlink it and free tree handle
     */
    if (!tree->next)
         tree->con->first_tree = tree->prev;
    else tree->next->prev      = tree->prev;

    if (!tree->prev)
         tree->con->last_tree = tree->next;
    else tree->prev->next     = tree->next;
  }
  RFCNB_Free_Pkt (pkt);
  return (0);
}


/*
 * Print some debug
 */
void SMB_debug (int level, const char *fmt, ...)
{
  if (SMBlib_debug >= level)
  {
    va_list arg;
    va_start (arg, fmt);
    vfprintf (stderr, fmt, arg);
    va_end (arg);
  }
}


